import java.rmi.*;
import java.rmi.server.UnicastRemoteObject;
import java.rmi.registry.*;
import java.net.*;
import java.util.*;

public class Client extends UnicastRemoteObject implements Client_itf {
	
	//The server
	private static Server_itf server;
	
	//The instance of the client
	private static Client_itf client;
	
	// SharedObjects of the Client (access by ID)
	protected static HashMap<Integer,SharedObject> idObjects;

	public Client() throws RemoteException {
		super();		
	}


////////////////////////////////////////////////////
//         Interface to be used by applications
////////////////////////////////////////////////////

	// initialization of the client layer
	public static void init() {
		try {
			Client.idObjects = new HashMap<Integer,SharedObject>();
			// Retrieves the server (remote)
			Client.server = (Server_itf)Naming.lookup("//localhost"+"/Server");
			// Initiates the object Client
			Client.client = new Client();
		} catch (Exception e) {
			e.printStackTrace();
		}
	}
	
	// lookup in the name server
	public static SharedObject lookup(String name) {
		SharedObject so = null;
		try {
			int id = Client.server.lookup(name);
			if (id != -1) { 
				//The object is already registered
				so = new SharedObject(id);
				System.out.println("Object succesfully retrieved. " +
						"Adding SharedObject");
				Client.idObjects.put(id,so);						
			}
		} catch (Exception e) {
			e.printStackTrace();
		}	
		return so;
	}		
	
	// binding in the name server
	public static void register(String name, SharedObject_itf so) {
		try {
			Client.server.register(name, ((SharedObject)so).getId());
		} catch (Exception e) {
			e.printStackTrace();
		}			
	}

	// creation of a shared object
	public static SharedObject create(Object o) {
		SharedObject so = null;
		try {
			//Ask the server for a new ID
			int id = Client.server.create(o);
			//Storage of the new SharedObject in the Client
			so = new SharedObject(id);
			Client.idObjects.put(id,so);
		} catch (Exception e) {
			e.printStackTrace();
		}	
		return so;
	}
	
/////////////////////////////////////////////////////////////
//    Interface to be used by the consistency protocol
////////////////////////////////////////////////////////////

	// request a read lock from the server
	public static Object lock_read(int id) {
		System.out.println("Appel à lock_read sur le Client");
		Object result = null;
		try {
			result = Client.server.lock_read(id, Client.client);
			if (result == null){
				System.out.println("Null for server.lock_read");
			}
			
			SharedObject so = Client.idObjects.get(id);
			
			so.obj = result;
		} catch (Exception e) {
			e.printStackTrace();
		}	
		return result;
	}

	// request a write lock from the server
	public static Object lock_write (int id) {
		System.out.println("Appel à lock_write sur le Client");
		Object result = null;
		try {
			result = Client.server.lock_write(id, Client.client);
			SharedObject so = Client.idObjects.get(id);
			so.obj = result;			
		} catch (Exception e) {
			e.printStackTrace();
		}	
		return result;
	}

	// receive a lock reduction request from the server
	public Object reduce_lock(int id) throws java.rmi.RemoteException {
		SharedObject so = Client.idObjects.get(id);
		Object result = so.reduce_lock();
		System.out.println("Leaving client reduce_lock");
		return result;
	}


	// receive a reader invalidation request from the server
	public void invalidate_reader(int id) throws java.rmi.RemoteException {
		SharedObject so = Client.idObjects.get(id);
		so.invalidate_reader();
	}


	// receive a writer invalidation request from the server
	public Object invalidate_writer(int id) throws java.rmi.RemoteException {
		SharedObject so = Client.idObjects.get(id);
		return so.invalidate_writer();
	}
}
