import java.lang.reflect.*;
import java.io.*;

public class Generator {
	
	public static void create_itfAndStub(String name){
		
		
		Class classe;
        try {
            classe = Class.forName(name);
        } catch (ClassNotFoundException e) {
            System.err.println ("Class "+name+" not found.");
            return;
        }
		
		String itfName = (String)(name + "_itf");
		
		String stubName = (String)(name + "_stub");
		
		PrintWriter pw_itf;
		
		PrintWriter pw_stub;
		
		
		try {
			pw_itf = new PrintWriter(new FileWriter(itfName + ".java",false));
		} catch (Exception e){
			System.out.println("Error creating interface file for : "+ itfName + ".java");
			e.printStackTrace();
			return;
		}
		
		
		try {
			pw_stub = new PrintWriter(new FileWriter(stubName + ".java",false));
		} catch (Exception e){
			System.out.println("Error creating stub file for : "+ stubName + ".java");
			e.printStackTrace();
			return;
		}
		
		//interface
		String classeType = "";
		if (classe.isInterface()) {
           classeType = "interface";
        } else {
            classeType = "class";
        }
		
		//write files
		pw_itf.println("public interface " + itfName + " extends SharedObject_itf { \n");
		
		pw_stub.println("public " + classeType +" "+ stubName + " extends SharedObject " +
				"implements " + itfName + ", java.io.Serializable { \n");
		
		
		//methods
		Method[] methodes = classe.getMethods();
		
		for (int i = 0; i < methodes.length; i++) {
	            Method meth = methodes[i];
	            
	            String methodName = meth.getName();
	            
	            if (   !methodName.equals("hashCode")
	                && !methodName.equals("getClass")
	                && !methodName.equals("wait")
	                && !methodName.equals("equals")
	                && !methodName.equals("notify")
	                && !methodName.equals("notifyAll")
	                && !methodName.equals("toString")         
	                ){
	            	
	            	Class[] params = meth.getParameterTypes();
	            	
	            	String returnType = meth.getReturnType().getName();
	            	
	            	pw_itf.print("\tpublic "+ returnType + " " +methodName
	            			+"(");
	            	
	            	pw_stub.print("\tpublic "+ returnType + " " +methodName
	            			+"(");
	            
		      
		            for (int j = 0; j < params.length; j++) {
		            	
		            	pw_itf.print(params[j].getName()+" param"+j);
		            	pw_stub.print(params[j].getName()+" param"+j);
		            	if (j!=params.length-1){
		            		pw_itf.print(",");
		            		pw_stub.print(",");
		            	}
		            }
		            
		            //for interface
		            pw_itf.println(");\n");
		            
		            //additional for stub file
		            pw_stub.println("){");
		            pw_stub.println("\t\t"+name + " s = (" 
		            		+ name + ")obj;");
		            
		            String ret = "";
		            if (!returnType.equals("void")){
		            	ret = "return";
		            }
		            pw_stub.print("\t\t" +
		            		ret+" s."+methodName+"(");
		            
		            for (int j = 0; j < params.length; j++) {
		            	pw_stub.print("param"+j);
		            	if (j!=params.length-1){
		            		pw_stub.print(",");
		            	} 
		            }
		            
		            pw_stub.print(");");
		            pw_stub.println("\n\t}\n");
		            
	            }
	        }
		
		
		
		//Closing itf main {
		pw_itf.println("}");
		pw_stub.println("}");
		
		pw_itf.close();
		pw_stub.close();
		
	}
	

	
	public static void main (String[] args) {
	
		if (args.length != 1) {
            System.err.println("Missing class or interface name.");
            return;
        }
		
		String classname = args[0];
		
		create_itfAndStub(classname);
		
	}

}
