import java.util.*;

public class ServerObject {
	
	//Same ID as the SharedObject
	private int id;
	
	//The shared Object
	private Object obj;
	
	//The list of reading clients
	private Vector<Client_itf> readingClients;

	//The writer (if any) (in WLC or WLT)
	private Client_itf writer;
	
	//State
	// 0 free
	// 1 someone is reading
	// 2 write mode
	public int state;

	public ServerObject(int i, Object o) {
		this.id = i;
		this.obj = o;
		this.readingClients = new Vector<Client_itf>();
		this.writer = null;
	}
	
	public int getId(){
		return this.id;
	}
	
	public Object getObject(){
		return this.obj;
	}
	
	public void setObject(Object o){
		this.obj = o;
	}
	
	public void addReader(Client_itf c){
		this.readingClients.add(c);
	}
	
	public void removeReader(Client_itf c){
		this.readingClients.remove(c);
	}
	
	public void setWriter(Client_itf c){
		this.writer = c;
	}
	
	public Client_itf getWriter(){
		return this.writer;
	}
	
	public int getState() {
		return this.state;
	}
	
	public synchronized Object lock_read(Client_itf client){
		System.out.println("Lock_read on serverObject : " + this.id);
		if (this.writer != null) {
			try {
				System.out.println("Reducing lock...");
				Object ob = this.writer.reduce_lock(this.id);
				if (ob != null) {
					this.obj = ob;
				}
				this.readingClients.add(this.writer);
				System.out.println("Writer has been lock-reduced");
			} catch (Exception e){
				e.printStackTrace();
			}
		} else {
			System.out.println("No one was writing");
		}		
		this.writer=null;
		System.out.println("Adding reader");
		this.readingClients.add(client);

		return this.obj;
		
	}
	
	public synchronized Object lock_write(Client_itf client){
		System.out.println("Lock_write on serverObject : " + this.id);
		if (this.writer != null){
			try {
				System.out.println("Invalidating writer...");
				Object ob = this.writer.invalidate_writer(this.id);
				if (ob != null) {
					this.obj=ob;
				}
				System.out.println("Writer has been invalidated");
			} catch (Exception e) {
				e.printStackTrace();
			}
		} else {
			System.out.println("No one was writing");
		}
		this.writer = null;
		
		Iterator<Client_itf> it = this.readingClients.iterator();
		while (it.hasNext()){
			try {
				System.out.println("Invalidating reader...");
				Client_itf cl = it.next();
				if (!cl.equals(client)) {
					cl.invalidate_reader(this.id);
				}
				System.out.println("done!");
			} catch (Exception e){
				e.printStackTrace();
			}
		}
		
		this.readingClients = new Vector<Client_itf>();
		this.writer = client;
		
		
		return this.obj;
	}

}
