import java.io.*;

public class SharedObject implements Serializable, SharedObject_itf {
	
	
	// State of the SharedObject
	// 0 NL       no local lock
	// 1 RLC      read lock cached (not taken)
	// 2 WLC      write lock cached
	// 3 RLT      read lock taken
	// 4 WLT      write lock taken
	// 5 RLT_WLC  read lock taken & write lock cached
	private int lock;
	
	// ID number
	private int id;
	
	// Shared object
	public Object obj;
	
	public SharedObject(){
		this.lock = State.NL;
		this.id = -1;
		this.obj = null;
		System.out.println("Création : " + this.lock);
	}
		
	public SharedObject(int i) {
		this.lock = State.NL;
		this.id = i;
		this.obj = null;
		System.out.println("Création : " + this.lock);
	}
	
	public void setId(int i){
		this.id = i;
	}
	
	public int getId(){
		return this.id;
	}
	
	public void setLock(int l){
		this.lock = l;
	}
	
	public int getLock(){
		return this.lock;
	}
	
	// invoked by the user program on the client node
	public void lock_read() {
		
		synchronized(this) {
		System.out.println("lock_read requested by user");
		System.out.println("     " + this.lock);
		
		boolean askClient = false;
		
		switch (this.lock) {
		
			case (State.RLC) : 
				this.lock = State.RLT; 
				break;
				
			// RLT -> impossible
				
			case (State.WLC) :
				this.lock = State.RLT_WLC;
				break;
				
			// WLT -> impossible
		    // RLT_WLC -> impossible
				
			case (State.NL) : 
				askClient = true;
				break;
				
			default :
				System.out.println("lock_read impossible : le SharedObject " +
						"est encore en lecture ou écriture" + this.lock);
			System.exit(0);
				break;
		
		}
		
		if (!askClient){ 
			System.out.println("   ->" + this.lock);
			System.out.println("Got lock_read !");
			return;
		}
		
		//So we must ask the client
		//this.lock=State.NL;
		}
		
		// ** Not synchronized **
		
		while (this.lock != State.RLT) {
		
			this.obj=Client.lock_read(this.id);
			
			synchronized(this){
				if (this.obj != null){
					this.lock = State.RLT;
				}
				//else the sharedObject just got
				// an invalidation from the server.
				// We must try again to lock_read 
			}
		
		}
		
		
		if (this.obj == null) {
			System.out.println("!!!! null in lock_read");
		}
		

		System.out.println("   ->" + this.lock);
		System.out.println("Got lock_read !");
		
	}

	// invoked by the user program on the client node
	public void lock_write() {	
		
		synchronized(this) {
			
		System.out.println("lock_write requested by user");
		System.out.println("     " + this.lock);
		
		boolean askClient = false;
		
		switch (this.lock) {
				
				case (State.RLC) : 
				case (State.NL) :	
					askClient = true;
					break;
							
				// RLT -> impossible
					
				case (State.WLC) : // WLC
					this.lock = State.WLT;
					break;
					
				// WLT -> impossible
					
				// RLT_WLC -> impossible				
				
				default :
					System.out.println("lock_write impossible : le SharedObject " +
							"est encore en lecture ou écriture : " + this.lock);
				System.exit(0);	
				break;
		}
		
		
		
		if (!askClient){ //WLC
			System.out.println("   ->" + this.lock);
			System.out.println("Got lock_write !");
			return;
		}
		
		//So we must ask the client
		
		//We unlock the SharedObject to avoid dead-lock ??
		//this.lock = State.NL;
		
		}
		
		// ** Not synchronized **
		
		while (this.lock != State.WLT) {

			this.obj=Client.lock_write(this.id);
			
			//now we can be invalidated (this.obj becomes null)
			synchronized(this){
				if (this.obj != null){
					this.lock = State.WLT;
				}
				// else the sharedObject just got
				// an invalidation from the server.
				// We must try again to lock_write
			}
		}
		
		if (this.obj == null) {
			System.out.println("!!!!!! null leaving lock_write");
			System.out.println("!!!!!!");
		}
		
		System.out.println("   ->" + this.lock);
		System.out.println("Got lock_write !");
		
		
	}

	// invoked by the user program on the client node
	public synchronized void unlock() {
		System.out.print("Unlocking...");
		switch (this.lock) {
		
		case (State.RLT) :
			this.lock = State.RLC;
			break;
		
		case (State.WLT) :
			this.lock = State.WLC;
			break;
		
		case (State.RLT_WLC) :
			this.lock = State.WLC;
			break;
			
		default: 
			System.out.println("Erreur : le SharedObject était " +
				"déjà déverrouillé");
		System.exit(0);
			break;
			
		}
		System.out.println(" new : "+this.lock);
		this.notify();
	}


	// callback invoked remotely by the server
	public synchronized Object reduce_lock() {
		System.out.println("Request by server : reduce_lock");
		switch (this.lock) {
		
		case (State.WLC) :
			this.lock = State.RLC;
			break;
			
		case (State.RLT_WLC) :
			this.lock = State.RLT;
			break;
			
		case (State.WLT) :
			while (this.lock == State.WLT) {
				// wait for the lock to become unlocked
				try {
					System.out.println("reduce_lock : waiting (WLT)");
					this.wait();
				} catch (Exception e) {
					e.printStackTrace();
				}
			}
		    if (this.lock == State.RLT || this.lock == State.RLT_WLC){
		    	System.out.println("Writer finished -> RLT (reduce_lock)");
				this.lock = State.RLT;			    	
		    }
		    if (this.lock == State.WLC){
			System.out.println("Writer finished -> RLC (reduce_lock)");
			this.lock = State.RLC;
		    }
			break;
			
		 	
		case (State.RLC) :
		case (State.NL) ://the writer doesn't know yet it is a writer
	          //so we will wait for the unlock anyway
			
			while (this.lock != State.WLC) {			
				try {
					this.wait();
				} catch (Exception e) {
					e.printStackTrace();
				}
			}
		System.out.println("Writer finished -> RLC (reduce_lock)");
			this.lock = State.RLC;
			
			
			break;
			
		default: break;
		
		}
		
		return this.obj;
	}

	// callback invoked remotely by the server
	public synchronized void invalidate_reader() {
		System.out.println("Request by server : invalidate_reader");
		System.out.println("Etat : "+this.lock);
		switch (this.lock) {
		
		case (State.RLC) :
		case (State.WLC) :
			this.lock = State.NL;
			break;
		
		case (State.RLT) :
		case (State.WLT) :
		case (State.RLT_WLC) :
			while (this.lock == State.RLT || this.lock == State.WLT || this.lock == State.RLT_WLC) {
				// wait for the lock to become NL
				try {
					System.out.println("invalidate_reader waiting : "+this.lock);
					this.wait();
				} catch (Exception e) {
					e.printStackTrace();
				}
			}
			// Always NL
			System.out.println("Reader finished -> NL (invalidate reader)");
			this.lock = State.NL;
		    break;
		    
		case (State.NL) : //the reader doesn't know yet it is a reader
	          			  //so we will wait for the unlock anyway
			while (this.lock != State.RLC){
				try {
					this.wait();
				} catch (Exception e) {
					e.printStackTrace();
				}
			}
			System.out.println("Reader finished -> NL (invalidate reader)");
			this.lock = State.NL;
			break;
			
		default : // useless
			System.out.println("ERREUR : invalidate_reader sur un NL");
		System.exit(0);
			break;	
			
		
		}
		System.out.println("Setting the object to NULL (invalidate reader)(new " +
				+this.lock+")");
		this.obj = null;
	}

	public synchronized Object invalidate_writer() {
		//We suppose the SharedObject becomes no lock
		// even if the SharedObject is RLT_WLC
		
		System.out.println("Request by server : invalidate_writer)");
		System.out.println("Etat : "+this.lock);
		
		switch (this.lock) {
		
		
		case (State.RLT) :			
			
			System.out.println("ERREUR : invalidate_writer sur un SO " +
					"en train de lire");
		System.exit(0);
			break;
			
		case (State.WLC) :
			//this.lock = State.RLC;
			this.lock = State.NL;
			break;
			
		case (State.WLT) :
			while (this.lock != State.WLC) {
				// wait for the lock to become NL
				try {
					System.out.println("invalidate_writer waiting (WLT)");
					this.wait();
				} catch (Exception e) {
					e.printStackTrace();
				}
			}
			//
		   System.out.println("Writer finished -> NL (invalidate writer)");
			this.lock = State.NL;
			break;
		
		case (State.RLT_WLC) :
			//this.lock = State.RLT;
			while (this.lock != State.WLC) {
				// wait for the lock to become NL
				try {
					System.out.println("invalidate_writer waiting (RLT_WLC)");
					this.wait();
				} catch (Exception e) {
					e.printStackTrace();
				}
			}
			System.out.println("Reader finished -> NL (invalidate writer)");
			this.lock = State.NL;
			break;
			
		case (State.RLC) :	
		case (State.NL) : //the writer doesn't know yet it is a writer
				          //so we will wait for the unlock anyway
			while (this.lock != State.WLC){
				try {
					this.wait();
				} catch (Exception e) {
					e.printStackTrace();
				}
			}
		    System.out.println("Writer finished -> NL (invalidate writer)");
			this.lock = State.NL;
			break;
			
		default: // useless
			System.out.println("ERREUR : invalidate_writer sur un NL");
		System.exit(0);
			break;
		}
		
		Object result = this.obj;
		System.out.println("Setting the object to NULL (invalidate writer)(new " +
				+this.lock+")");
		this.obj = null;
		return result;		
	}
}
